#include <linux/module.h>
#include <linux/init.h>
#include <linux/kernel.h>
#include <linux/fs.h>
#include <linux/cdev.h>
#include <asm/uaccess.h>
#include <linux/sched.h>
#include <linux/wait.h>

MODULE_DESCRIPTION("Read string from device");
MODULE_AUTHOR("SO2");
MODULE_LICENSE("GPL");

#define LOG_LEVEL		KERN_DEBUG

#define MY_MAJOR		43
#define MY_MINOR		0
#define NUM_MINORS		1
#define MODULE_NAME		"readstring"
#define READ_MESSAGE		"anaaremere"

#define BUFFER_SIZE		256

struct device_data_struct {
	struct cdev cdev;
	char buffer[BUFFER_SIZE];
	size_t buffer_len;
} devs[NUM_MINORS];

static int print_open(struct inode *inode, struct file *file)
{
	struct device_data_struct *data =
		container_of(inode->i_cdev, struct device_data_struct, cdev);

	file->private_data = data;

	return 0;
}

static int print_release(struct inode *inode, struct file *file)
{
	return 0;
}

static ssize_t string_read(struct file *file,
		char __user *user_buffer,
		size_t size, loff_t *offset)
{
	struct device_data_struct *data = file->private_data;
	size_t to_read;

	printk(LOG_LEVEL "Calling read for %zu bytes\n", size);

	if (size > data->buffer_len - *offset)
		to_read = data->buffer_len - *offset;
	else
		to_read = size;
	if (copy_to_user(user_buffer, data->buffer + *offset, to_read) != 0)
		return -EFAULT;
	*offset += to_read;

	printk(LOG_LEVEL "Return %zu bytes\n", to_read);

	return to_read;
}

static const struct file_operations print_fops = {
	.owner = THIS_MODULE,
	.open = print_open,
	.release = print_release,
	.read = string_read,
};

static int string_init(void)
{
	int err;
	int i;

	err = register_chrdev_region(MKDEV(MY_MAJOR, MY_MINOR),
			NUM_MINORS, MODULE_NAME);
	if (err != 0) {
		printk(KERN_ERR "register_chrdev_region");
		return err;
	}

	for (i = 0; i < NUM_MINORS; i++) {
		cdev_init(&devs[i].cdev, &print_fops);
		cdev_add(&devs[i].cdev, MKDEV(MY_MAJOR, i), 1);
		strcpy(devs[i].buffer, READ_MESSAGE);
		devs[i].buffer_len = strlen(devs[i].buffer);
	}

	return 0;
}

static void string_exit(void)
{
	int i;

	for (i = 0; i < NUM_MINORS; i++)
		cdev_del(&devs[i].cdev);

	unregister_chrdev_region(MKDEV(MY_MAJOR, MY_MINOR), NUM_MINORS);
}

module_init(string_init);
module_exit(string_exit);
